// Bookmark Backup
// By Pike

// Contributor(s):
// Nickolay Ponomarev (Multiple file backup code)

function BackupService()
{
	this._init();
}

BackupService.prototype = {

	_GROUPS_URL: 'chrome://bookmarkbackup/content/groups.js',
	_STRINGS_URL: 'chrome://bookmarkbackup/locale/strings.properties',
	_CREATE_PERMISSIONS: 0700,
	
	performBackup: function()
	{
		// Create logger
		this._logger = new Logger();
		this._logger.log('Backup begun');
		try
		{
			// Init vars
			var profileDir = this._dirService.get('ProfD', Components.interfaces.nsILocalFile);
			var stringBundleService = Components.classes['@mozilla.org/intl/stringbundle;1'].getService(Components.interfaces.nsIStringBundleService);
			var localeStrings = stringBundleService.createBundle(this._STRINGS_URL);
			
			// Find bookmarks file and backups dir
			var bookmarksFile = this._getBookmarksFile();
			var backupsDirPath = localeStrings.GetStringFromName('backupsFolderName');
			var dayNameDirPath = localeStrings.GetStringFromName('dayOfWeek.' + (new Date()).getDay());
			var backupsDir = this._getBackupsDir(backupsDirPath, dayNameDirPath, bookmarksFile);
			
			if (backupsDir)
			{
				// Get names of additional files to backup
				var backupFileNames = [];
				var chosenGroups = this._prefBranch.getComplexValue('bookmarkbackup.chosenGroups', Components.interfaces.nsISupportsString).data;
				chosenGroups = chosenGroups.split('|');
				chosenGroups.unshift('_default'); // Add default group to start of array
				for (var i = 0; i < chosenGroups.length; i++)
				{
					var group = this.getBackupGroup(chosenGroups[i], {});
					if (group)
					{
						backupFileNames = backupFileNames.concat(group);
					}
				}
				var chosenFiles = this._prefBranch.getComplexValue('bookmarkbackup.chosenFiles', Components.interfaces.nsISupportsString).data;
				backupFileNames = backupFileNames.concat(chosenFiles.split('|'));
				
				// Backup bookmarks file
				if (bookmarksFile)
				{
					this._copyFile(bookmarksFile.parent, backupsDir.clone(), [bookmarksFile.leafName]);
				}
				
				// Backup additional files
				for (var j = 0; j < backupFileNames.length; j++)
				{
					if (backupFileNames[j])
					{
						var fileSections = backupFileNames[j].split('/');
						this._copyFile(profileDir.clone(), backupsDir.clone(), fileSections);
					}
				}
				
				this._logger.log('Backup completed');
			}
		}
		catch (err)
		{
			if (err instanceof BackupFolderNotCreated)
			{
				this._logger.log('Error: Today\'s backup dir could not be created at: ' + err.path);
			}
			else
			{
				this._logger.log('Error: Backup aborted due to ' + err.name + ' at line ' + err.lineNumber + ':\n' + err.message);
			}
		}
		this._logger.addBlankLine();
		this._logger = null;
	},
	
	get backupGroups()
	{
		var groupsData = (typeof BackupGroups == 'object' ? BackupGroups : null);
		return new GroupsEnumerator(groupsData);
	},
	
	getBackupGroup: function(id, count)
	{
		if (typeof BackupGroups == 'object' && BackupGroups[id])
		{
			count.value = BackupGroups[id].length;
			return BackupGroups[id];
		}
		else
		{
			count.value = 0;
			return null;
		}
	},
	
	_getBookmarksFile: function()
	{
		var bookmarksFile;
		try
		{
			// As of Fx 1.5 the BMarks key *does* take the browser.bookmarks.file pref into account
			bookmarksFile = this._dirService.get('BMarks', Components.interfaces.nsILocalFile);
		}
		catch (err)
		{
			bookmarksFile = null;
		}
		return bookmarksFile;
	},
	
	_getStandardBackupsDir: function(backupsDirName, dayNameDirPath, bookmarksFile)
	{
		var standardBackupsDir;
		if (bookmarksFile) // Use directory in same place as bookmarks.html
		{
			standardBackupsDir = bookmarksFile.parent;
		}
		else // Use directory in profile
		{
			standardBackupsDir = this._dirService.get('ProfD', Components.interfaces.nsILocalFile);
		}
		standardBackupsDir.append(backupsDirName);
		standardBackupsDir.append(dayNameDirPath);
		return standardBackupsDir;
	},
	
	_getCustomBackupsDir: function(dayNameDirPath)
	{
		var customBackupsDir;
		try
		{
			customBackupsDir = this._prefBranch.getComplexValue('bookmarkbackup.location', Components.interfaces.nsILocalFile);
			customBackupsDir.append(dayNameDirPath);
		}
		catch (err)
		{
			customBackupsDir = null;
		}
		return customBackupsDir;
	},
	
	_getBackupsDir: function(backupsDirName, dayNameDirPath, bookmarksFile)
	{
		var deletePrevious = this._prefBranch.getBoolPref('bookmarkbackup.deletePrevious');
		
		// If a custom location was set, try to use it
		var backupsDir = this._getCustomBackupsDir(dayNameDirPath);
		if (backupsDir)
		{
			if (deletePrevious && backupsDir.exists())
			{
				backupsDir.remove(true);
				this._logger.log('Today\'s backup dir deleted at: ' + backupsDir.path);
			}
			
			if (!backupsDir.exists())
			{
				try
				{
					// Create backup dir if it doesn't exist yet
					backupsDir.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, this._CREATE_PERMISSIONS);
					this._logger.log('Today\'s backup dir created at: ' + backupsDir.path);
				}
				catch (err)
				{
					// Custom dir could not be used, use standard location instead
					backupsDir = null;
					this._logger.log('Warning: Custom backup location was invalid, using default location instead (usually the user profile)');
				}
			}
			else
			{
				this._logger.log('Today\'s backup dir found at: ' + backupsDir.path);
			}
		}
		
		// If no custom location was set, or custom location was invalid, get standard location
		if (!backupsDir)
		{
			backupsDir = this._getStandardBackupsDir(backupsDirName, dayNameDirPath, bookmarksFile);
			if (backupsDir)
			{
				if (deletePrevious && backupsDir.exists())
				{
					backupsDir.remove(true);
					this._logger.log('Today\'s backup dir deleted at: ' + backupsDir.path);
				}
				
				if (!backupsDir.exists())
				{
					try
					{
						// Create backup dir if it doesn't exist yet
						backupsDir.create(Components.interfaces.nsIFile.DIRECTORY_TYPE, this._CREATE_PERMISSIONS);
						this._logger.log('Today\'s backup dir created at: ' + backupsDir.path);
					}
					catch (err)
					{
						throw new BackupFolderNotCreated(backupsDir.path);
					}
				}
				else
				{
					this._logger.log('Today\'s backup dir found at: ' + backupsDir.path);
				}
			}
		}
		
		return backupsDir;
	},
	
	_copyFile: function(fromFileOrDir, toFileOrDir, relPath)
	// Perform recursive backup. An alternative would be using nsIFile.remove(true) on the existing
	// backup, but this function is safer as (in theory) it should only remove files that are about
	// to be overwritten, parameter types are: (nsIFile, nsIFile, string array)
	{
		if (fromFileOrDir.exists())
		{
			if (relPath.length > 0)
			{
				var pathSection = relPath.shift();
				var wildcard = /^\*\.(\w*)$/i.exec(pathSection);
				if (wildcard)
				{
					this._copyChildren(fromFileOrDir, toFileOrDir, relPath, wildcard[1]);
				}
				else
				{
					fromFileOrDir.append(pathSection);
					toFileOrDir.append(pathSection);
					this._copyFile(fromFileOrDir, toFileOrDir, relPath);
				}
			}
			else if (fromFileOrDir.isDirectory())
			{
				this._copyChildren(fromFileOrDir, toFileOrDir, relPath, null);
			}
			else
			{
				try
				{
					if (toFileOrDir.exists())
					{
						toFileOrDir.remove(false);
					}
					fromFileOrDir.copyTo(toFileOrDir.parent, toFileOrDir.leafName);
					this._logger.log(toFileOrDir.leafName + ' successfully copied');
				}
				catch (err)
				{
					this._logger.log('Warning: ' + fromFileOrDir.leafName + ' could not be copied to ' + toFileOrDir.path);
				}
			}
		}
		else
		{
			this._logger.log('Warning: ' + fromFileOrDir.leafName + ' could not be found');
		}
	},
	
	_copyChildren: function(fromDir, toDir, relPath, fileExt)
	{
		// Check fromDir exists and is a directory, DO NOT check toDir since it may not exist yet
		if (fromDir.exists() && fromDir.isDirectory())
		{
			var pattern = null;
			if (fileExt)
			{
				pattern = new RegExp('\\.' + fileExt + '$');
			}
			
			var dirEnum = fromDir.directoryEntries;
			while (dirEnum.hasMoreElements())
			{
				var fromChild = dirEnum.getNext().QueryInterface(Components.interfaces.nsIFile);
				if (!pattern || fromChild.leafName.match(pattern))
				{
					var toChild = toDir.clone();
					toChild.append(fromChild.leafName);
					this._copyFile(fromChild, toChild, relPath);
				}
			}
		}
	},
	
	_migratePrefs: function()
	{
		if (this._prefBranch.prefHasUserValue('bookmarkbackup.backupList'))
		{
			// Create hash table
			var hashTable = {};
			var backupGroups = this.backupGroups;
			while (backupGroups.hasMore())
			{
				var groupName = backupGroups.getNext();
				var group = this.getBackupGroup(groupName, {});
				for (var i = 0; i < group.length; i++)
				{
					if (!hashTable[group[i]])
					{
						hashTable[group[i]] = [];
					}
					hashTable[group[i]].push(groupName);
				}
			}
			
			// Assign each file in list to a backup group
			var chosenGroups = [], chosenFiles = [];
			var backupList = this._prefBranch.getComplexValue('bookmarkbackup.backupList', Components.interfaces.nsISupportsString).data.split('|');
			for (var j = 0; j < backupList.length; j++)
			{
				if (hashTable[backupList[j]])
				{
					for (var k = 0; k < hashTable[backupList[j]].length; k++)
					{
						chosenGroups.push(hashTable[backupList[j]][k]);
					}
				}
				else
				{
					chosenFiles.push(backupList[j]);
				}
			}
			
			// Set prefs
			this._setArrayPref(chosenGroups, 'bookmarkbackup.chosenGroups');
			this._setArrayPref(chosenFiles, 'bookmarkbackup.chosenFiles');
			this._prefBranch.clearUserPref('bookmarkbackup.backupList');
		}
	},
	
	// XXX - FIXME: This method is duplicated in both the component and the settings code
	_setArrayPref: function(list, prefName)
	{
		// Remove duplicates
		var prunedList = [];
		var lastValue;
		list.sort();
		for (var i = 0; i < list.length; i++)
		{
			if (!i || list[i] != lastValue)
			{
				prunedList.push(list[i]);
				lastValue = list[i];
			}
		}
		
		// Save pref
		if (prunedList.length)
		{
			var listStr = Components.classes['@mozilla.org/supports-string;1'].createInstance(Components.interfaces.nsISupportsString);
			listStr.data = prunedList.join('|');
			this._prefBranch.setComplexValue(prefName, Components.interfaces.nsISupportsString, listStr);
		}
		else
		{
			try
			{
				this._prefBranch.clearUserPref(prefName);
			}
			catch (err) {}
		}
	},
	
	_init: function()
	{
		// Setup object properties
		this._observerService = Components.classes['@mozilla.org/observer-service;1'].getService(Components.interfaces.nsIObserverService);
		this._dirService = Components.classes['@mozilla.org/file/directory_service;1'].getService(Components.interfaces.nsIProperties);
		this._prefBranch = Components.classes['@mozilla.org/preferences-service;1'].getService(Components.interfaces.nsIPrefBranch);
		this._timerHandle = Components.classes['@mozilla.org/timer;1'].createInstance(Components.interfaces.nsITimer);
		
		// Add observers
		this._observerService.addObserver(this, 'profile-after-change', false);
		this._observerService.addObserver(this, 'profile-before-change', false);
		this._observerService.addObserver(this, 'xpcom-shutdown', false);
	},
	
	_destruct: function()
	{
		// Remove observers
		this._observerService.removeObserver(this, 'profile-after-change');
		this._observerService.removeObserver(this, 'profile-before-change');
		this._observerService.removeObserver(this, 'xpcom-shutdown');
		
		// Clear object properties
		this._observerService = null;
		this._dirService = null;
		this._prefBranch = null;
		this._timerHandle = null;
	},
	
	_onLoad: function()
	{
		// Load backup groups
		try
		{
			var jsSubScriptLoader = Components.classes['@mozilla.org/moz/jssubscript-loader;1'].getService(Components.interfaces.mozIJSSubScriptLoader);
			jsSubScriptLoader.loadSubScript(this._GROUPS_URL);
		}
		catch (err) {}
		
		// Migrate old prefs
		this._migratePrefs();
		
		// Perform initial backup
		if (this._prefBranch.getBoolPref('bookmarkbackup.startupBackup'))
		{
			this.performBackup();
		}
		
		// Start cyclic backup
		var backupInterval = this._prefBranch.getIntPref('bookmarkbackup.interval');
		if (backupInterval > 0) // Setting the pref to 0 disables the cyclic backup
		{
			backupInterval *= 60000; // Convert from minutes to milliseconds
			this._timerHandle.init(this, backupInterval, Components.interfaces.nsITimer.TYPE_REPEATING_SLACK);
		}
	},
	
	_onUnload: function()
	{
		// Stop cyclic backup
		this._timerHandle.cancel();
		
		// Perform final backup
		if (this._prefBranch.getBoolPref('bookmarkbackup.shutdownBackup'))
		{
			this.performBackup();
		}
	},
	
	observe: function(subject, topic, data)
	{
		if (topic == 'profile-after-change' && data == 'startup')
		{
			this._onLoad();
		}
		else if (topic == 'profile-before-change' && data == 'shutdown-persist')
		{
			this._onUnload();
		}
		else if (topic == 'xpcom-shutdown')
		{
			this._destruct();
		}
		else if (topic == 'timer-callback' && subject == this._timerHandle)
		{
			this.performBackup();
		}
	},
	
	QueryInterface: function(iid)
	{
		if (iid.equals(Components.interfaces.iBackupService) ||
			iid.equals(Components.interfaces.nsIObserver) ||
			iid.equals(Components.interfaces.nsISupports))
		{
			return this;
		}
		else
		{
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
	}
};

function BackupFolderNotCreated(path)
{
	this._path = path;
}

BackupFolderNotCreated.prototype = {

	get path()
	{
		return this._path;
	}
};

function GroupsEnumerator(groups)
{
	this._index = 0;
	this._data = [];
	if (typeof groups == 'object')
	{
		for (var i in groups)
		{
			this._data.push(i);
		}
	}
}

GroupsEnumerator.prototype = {

	hasMore: function()
	{
		return this._index < this._data.length;
	},
	
	getNext: function()
	{
		return this._data[this._index++];
	},
	
	QueryInterface: function(iid)
	{
		if (iid.equals(Components.interfaces.nsIStringEnumerator) ||
			iid.equals(Components.interfaces.nsISupports))
		{
			return this;
		}
		else
		{
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
	}
};

function Logger()
{
	this._init();
}

Logger.prototype = {

	_CREATE_PERMISSIONS: 0600,
	_FILE_FLAGS: 0x02 | 0x08 | 0x10, // WRONLY | CREATE_FILE | APPEND
	
	log: function(message)
	{
		this._writeOutput((new Date().toLocaleString()) + ': ' + message + '\n');
	},
	
	addBlankLine: function()
	{
		this._writeOutput('\n');
	},
	
	_writeOutput: function(output)
	{
		if (this._logFile)
		{
			try
			{
				var fileOutputStream = Components.classes['@mozilla.org/network/file-output-stream;1'].createInstance(Components.interfaces.nsIFileOutputStream);
				fileOutputStream.init(this._logFile, this._FILE_FLAGS, this._CREATE_PERMISSIONS, 0);
				fileOutputStream.write(output, output.length);
				fileOutputStream.close();
			}
			catch (err) {}
		}
	},
	
	_init: function()
	{
		var prefBranch = Components.classes['@mozilla.org/preferences-service;1'].getService(Components.interfaces.nsIPrefBranch);
		try
		{
			if (prefBranch.getBoolPref('bookmarkbackup.loggingEnabled'))
			{
				var dirService = Components.classes['@mozilla.org/file/directory_service;1'].getService(Components.interfaces.nsIProperties);
				this._logFile = dirService.get('ProfD', Components.interfaces.nsILocalFile);
				this._logFile.append('bookmarkbackup.log');
			}
		}
		catch (err) {}
	}
};

var ModuleAndFactory = {

	NAME: 'BackupService@pikey.me.uk',
	
	CID: Components.ID('{a087a93d-0eea-4e10-958a-1a303a513ed5}'),
	
	CONTRACT_ID: '@pikey.me.uk/backup-service;1',
	
	IMPLEMENTATION: BackupService,
	
	registerSelf: function(compMgr, fileSpec, location, type)
	{
		compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
		compMgr.registerFactoryLocation(this.CID, this.NAME, this.CONTRACT_ID, fileSpec, location, type);
		
		var categoryManager = Components.classes['@mozilla.org/categorymanager;1'].getService(Components.interfaces.nsICategoryManager);
		categoryManager.addCategoryEntry('app-startup', this.NAME, 'service,' + this.CONTRACT_ID, true, true);
	},
	
	unregisterSelf: function(compMgr, fileSpec, location)
	{
		compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
		compMgr.unregisterFactoryLocation(this.CID, fileSpec);
		
		var categoryManager = Components.classes['@mozilla.org/categorymanager;1'].getService(Components.interfaces.nsICategoryManager);
		categoryManager.deleteCategoryEntry('app-startup', this.NAME, true);
	},
	
	getClassObject: function(compMgr, cid, iid)
	{
		if (!iid.equals(Components.interfaces.nsIFactory))
		{
			throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
		}
		
		if (cid.equals(this.CID))
		{
			return this;
		}
		else
		{
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
	},
	
	canUnload: function(compMgr)
	{
		return true;
	},
	
	createInstance: function(outer, iid)
	{
		if (outer)
		{
			throw Components.results.NS_ERROR_NO_AGGREGATION;
		}
		
		return (new this.IMPLEMENTATION()).QueryInterface(iid);
	}
};

function NSGetModule(compMgr, fileSpec)
{
	return ModuleAndFactory;
}
